from .baseresult import BaseResult
from .finalresult import FinalResult
import os

class StoredResult(BaseResult):

    def __init__(self, path, prefix="res"):
        
        self._path = path
        self._prefix = prefix
        BaseResult.__init__(self, self._path)
     
    def get_fp_file(self):
        return os.path.join(self._path, "%s_fp.csv" % self._prefix)

    def get_probtraj_file(self):
        return os.path.join(self._path, "%s_probtraj.csv" % self._prefix)

    def get_statdist_file(self):
        return os.path.join(self._path, "%s_statdist.csv" % self._prefix)


class StoredFinalResult(FinalResult):

    def __init__(self, path, prefix="res"):
        
        self._path = path
        self._prefix = prefix
        FinalResult.__init__(self, self._path)
     
    def get_last_states_file(self):
        return os.path.join(self._path, "%s_finalprob.csv" % self._prefix)



"""Function to import the ginsim out in Python.

The ginsim output is a special case of MaBoSS format. This parser uses
the very specific structure of the ginsim output and not able to parse every
MaBoSS file.
"""

from __future__ import print_function
import sys
from collections import OrderedDict
from sys import stderr, version_info
if version_info[0] < 3:
    from contextlib2 import ExitStack
else:
    from contextlib import ExitStack
from os.path import isfile
import pyparsing as pp
from .logic import varName
from .network import Node, Network
from .simulation import Simulation
from .cmabosssimulation import CMaBoSSSimulation
externVar = pp.Suppress('$') + ~pp.White() + varName
externVar.setParseAction(lambda token: token[0])
import uuid

# ====================
# bnd grammar
# ====================
internal_var_decl = pp.Group(varName('lhs') + pp.Suppress('=')
                             + pp.SkipTo(';')('rhs') + pp.Suppress(';'))
node_decl = pp.Group(pp.Suppress(pp.CaselessKeyword("Node")) + varName("name") + pp.Suppress('{')
                     + pp.OneOrMore(internal_var_decl)('interns')
                     + pp.Suppress('}'))
bnd_grammar = pp.OneOrMore(node_decl)
bnd_grammar.ignore('//' + pp.restOfLine)

# =====================
# cfg grammar
# =====================

intPart = pp.Word(pp.nums)
intPart.setParseAction(lambda token: int(token[0]))
floatNum = pp.Word(pp.nums + '.' + 'E' + 'e' + '-' + '+')
floatNum.setParseAction(lambda token: float(token[0]))
booleanStr = (pp.oneOf('0 1')
              | pp.CaselessLiteral("True") | pp.CaselessLiteral("False"))


def booleanStrAction(token):
    if pp.CaselessLiteral("True").matches(token) or pp.Word("1").matches(token):
        return True
    elif pp.CaselessLiteral("False").matches(token) or pp.Word("0").matches(token):
        return False
    else:
        print("Cannot recognize boolean value : %s" % token)
        return token


booleanStr.setParseAction(lambda token: booleanStrAction(token[0]))
numOrBool = (floatNum | booleanStr)("value")

var_decl = pp.Group(externVar("lhs") + pp.Suppress('=')
                    + pp.SkipTo(';')("rhs") + pp.Suppress(';'))


param_decl = pp.Group(varName("param") + pp.Suppress('=')
                      + numOrBool + pp.Suppress(';'))

stateSet = (pp.Suppress('[') + pp.Group(pp.delimitedList(intPart))
            + pp.Suppress(']'))
stateSet.setParseAction(lambda token: list(token))

stateProb = pp.Word(pp.alphanums+'()+-*/$.')('proba') + stateSet("states")
stateProb.setParseAction(lambda token: (token.proba, token.states))

istate_decl = pp.Group(pp.Suppress('[') + pp.delimitedList(varName)("nodes")
                       + pp.Suppress('].istate') + pp.Suppress('=')
                       + pp.delimitedList(stateProb)('attrib') + pp.Suppress(';'))

oneIstate_decl = pp.Group(varName("nd_i") + ~pp.White() + pp.Suppress('.istate')
                          + pp.Suppress('=') + booleanStr('istate_val')
                          + pp.Suppress(';'))

internal_decl = pp.Group(varName("node") + ~pp.White()
                         + pp.Suppress(".is_internal") + pp.Suppress('=')
                         + booleanStr("is_internal_val")
                         + pp.Suppress(';'))

refstate_decl = pp.Group(varName("node") + ~pp.White()
                         + pp.Suppress(".refstate") + pp.Suppress('=')
                         + numOrBool("refstate_val")
                         + pp.Suppress(';'))

cfg_decl = (var_decl | istate_decl | param_decl | internal_decl
            | oneIstate_decl | refstate_decl)
cfg_grammar = pp.ZeroOrMore(cfg_decl)
cfg_grammar.ignore('//' + pp.restOfLine)

def loadBNet(bnet_filename):
    assert bnet_filename.lower().endswith(".bnet"), "wrong extension for BNet file"

    if "://" in bnet_filename:
        from colomoto_jupyter.io import ensure_localfile
        bnet_filename = ensure_localfile(bnet_filename)

    from colomoto_jupyter import import_colomoto_tool
    biolqm = import_colomoto_tool("biolqm")
    return biolqm.to_maboss(biolqm.load(bnet_filename))

def load(bnd_filename, *cfg_filenames, **extra_args):
    """Loads a network from a MaBoSS format file.

    :param str bnd_filename: Network file
    :param str cfg_filename: Configuraton file
    :keyword str simulation_name: name of the returned :py:class:`.Simulation` object
    :rtype: :py:class:`.Simulation`
    """
    assert bnd_filename.lower().endswith(".bnd"), "wrong extension for bnd file"

    if "://" in bnd_filename:
        from colomoto_jupyter.io import ensure_localfile
        bnd_filename = ensure_localfile(bnd_filename)

    if not cfg_filenames: 
        cfg_filenames = [".".join([".".join(bnd_filename.split(".")[:-1]), "cfg"])]

    elif "://" in " ".join(cfg_filenames):
        from colomoto_jupyter.io import ensure_localfile
        cfg_filenames = [ensure_localfile(cfg_filename) if "://" in cfg_filename else cfg_filename for cfg_filename in cfg_filenames]

    if extra_args.get("cmaboss"):
        return CMaBoSSSimulation(bnd_filename, *cfg_filenames)

    command = extra_args.get("command")

    with ExitStack() as stack:
        bnd_file = stack.enter_context(open(bnd_filename, 'r'))
        bnd_content = bnd_file.read()

        cfg_content = ""
        for cfg_filename in tuple(cfg_filenames):
            cfg_file = stack.enter_context(open(cfg_filename, 'r'))
            cfg_content += cfg_file.read()

        (variables, parameters, is_internal_list,
         istate_list, refstate_list) = _read_cfg(cfg_content)

        nodes = _read_bnd(bnd_content, is_internal_list)

        net = Network(nodes)
        for istate in istate_list:
            net.set_istate(istate, istate_list[istate])
        for v in variables:
            lhs = '$'+v
            parameters[lhs] = variables[v]
        ret = Simulation(net, parameters, command=command)
        ret.refstate = refstate_list
        return ret


def _read_cfg(string):
        variables = OrderedDict()
        parameters = OrderedDict()
        is_internal_list = {}
        istate_list = OrderedDict()
        refstate_list = {}
        parse_cfg = cfg_grammar.parseString(string)
        for token in parse_cfg:
            if token.lhs:  # True if token is var_decl
                variables[token.lhs] = token.rhs
            if token.is_internal_val:  # True if token is internal_decl
                is_internal_list[token.node] = token.is_internal_val
            if token.refstate_val:
                refstate_list[token.node] = token.refstate_val
            if token.param:  # True if token is param_decl
                parameters[token.param] = float(token.value)
            if token.nd_i:
                istate_list[token.nd_i] = {0: 1 - int(token.istate_val),
                                           1: int(token.istate_val)}
            if token.attrib:  # True if token is istate_decl
                # TODO check if lens are consistent
                if len(token.nodes) == 1:
                    t_istate_list = OrderedDict()
                    for t in token.attrib:
                        try:
                            t_istate_list[int(t[1][0])] = float(str(t[0]))
                        except ValueError:
                            t_istate_list[int(t[1][0])] = str(t[0])

                    istate_list[token.nodes[0]] = t_istate_list

                else:
                    nodes = tuple(token.nodes)
                    t_istate_list = {}
                    for t in token.attrib:
                        try:
                            t_istate_list[tuple(t[1])] = float(str(t[0]))
                        except ValueError:
                            t_istate_list[tuple(t[1])] = str(t[0])

                    istate_list[nodes] = t_istate_list

        return (variables, parameters, is_internal_list, istate_list,
                refstate_list)


def _read_bnd(string, is_internal_list):
        nodes = []
        parse_bnd = bnd_grammar.parseString(string)

        for token in parse_bnd:
            interns = {v.lhs: v.rhs for v in token.interns}
            logic = interns.pop('logic') if 'logic' in interns else None
            rate_up = interns.pop('rate_up') if 'rate_up' in interns.keys() else "@logic ? 1.0 : 0.0"
            rate_down = interns.pop('rate_down') if 'rate_down' in interns.keys() else "@logic ? 0.0 : 1.0"

            internal = (is_internal_list[token.name]
                        if token.name in is_internal_list
                        else False)
            nodes.append(Node(token.name, logic, rate_up, rate_down,
                              internal, interns))
        return nodes



# coding: utf-8
# MaBoSS (Markov Boolean Stochastic Simulator)
# Copyright (C) 2011-2018 Institut Curie, 26 rue d'Ulm, Paris, France
   
# MaBoSS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
   
# MaBoSS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
   
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA 

# Module: maboss/comm.py
# Authors: Eric Viara <viara@sysra.com>
# Date: May-December 2018

from __future__ import print_function

from .result import Result
#
# MaBoSS Communication Layer
#

PROTOCOL_VERSION_NUMBER = "1.0"

MABOSS_MAGIC = "MaBoSS-2.0"
PROTOCOL_VERSION = "Protocol-Version:"
FLAGS = "Flags:"
HEXFLOAT_FLAG = 0x1
OVERRIDE_FLAG = 0x2
AUGMENT_FLAG = 0x4
COMMAND = "Command:"
RUN_COMMAND = "run"
CHECK_COMMAND = "check"
PARSE_COMMAND = "parse"
NETWORK = "Network:"
CONFIGURATION = "Configuration:"
CONFIGURATION_EXPRESSIONS = "Configuration-Expressions:"
CONFIGURATION_VARIABLES = "Configuration-Variables:"
RETURN = "RETURN"
STATUS = "Status:"
ERROR_MESSAGE = "Error-Message:"
STATIONARY_DISTRIBUTION = "Stationary-Distribution:"
TRAJECTORY_PROBABILITY = "Trajectory-Probability:"
TRAJECTORIES = "Trajectories:"
FIXED_POINTS = "Fixed-Points:"
RUN_LOG = "Run-Log:"

class HeaderItem:

    def __init__(self, directive, _from = None, to = None, value = None):
        self._directive = directive
        self._from = _from
        self._to = to
        self._value = value

    def getDirective(self):
        return self._directive

    def getFrom(self):
        return self._from

    def getTo(self):
        return self._to

    def getValue(self):
        return self._value

class DataStreamer:

    @staticmethod
    def buildStreamData(client_data, hints = None):
        data = ""
        offset = 0
        o_offset = 0

        if hints:
            hexfloat = "hexfloat" in hints and hints["hexfloat"]
            override = "override" in hints and hints["override"]
            augment = "augment" in hints and hints["augment"]
            verbose = "verbose" in hints and hints["verbose"]
        else:
            hexfloat = False
            override = False
            augment = False
            verbose = False

        command = client_data.getCommand()
        flags = 0
        if hexfloat:
            flags |= HEXFLOAT_FLAG
        if override:
            flags |= OVERRIDE_FLAG
        if augment:
            flags |= AUGMENT_FLAG

        header = MABOSS_MAGIC + "\n"
        header += PROTOCOL_VERSION + PROTOCOL_VERSION_NUMBER + "\n";
        header += FLAGS + str(flags) + "\n";
        header += COMMAND + command + "\n";

        config_data = client_data.getConfig()
        offset += len(config_data)
        data += config_data

        (header, o_offset) = DataStreamer._add_header(header, CONFIGURATION, o_offset, offset)

        network_data = client_data.getNetwork()
        offset += len(network_data)
        data += network_data

        (header, o_offset) = DataStreamer._add_header(header, NETWORK, o_offset, offset)

        if verbose:
            print("======= sending header\n", header)
            print("======= sending data[0:200]\n", data[0:200], "\n[...]\n")
        return header + "\n" + data

    @staticmethod
    def parseStreamData(ret_data, simulation, hints = None):
        verbose = False
        if hints:
            verbose = "verbose" in hints and hints["verbose"]

        result_data = Result(simulation)
        magic = RETURN + " " + MABOSS_MAGIC
        magic_len = len(magic)
        if ret_data[0:magic_len] != magic:
            result_data.setStatus(1)
            result_data.setErrorMessage("magic " + magic + " not found in header")
            return result_data

        offset = magic_len
        pos = ret_data.find("\n\n", magic_len)
        if pos < 0:
            result_data.setStatus(2)
            result_data.setErrorMessage("separator double nl found in header")
            return result_data

        offset += 1
        header = ret_data[offset:pos+1]
        data  = ret_data[pos+2:]
        if verbose:
            print("======= receiving header \n", header)
            print("======= receiving data[0:200]\n", data[0:200], "\n[...]\n")

        header_items = []
        err_data = DataStreamer._parse_header_items(header, header_items)
        if err_data:
            result_data.setStatus(3)
            result_data.setErrorMessage(err_data)
            return result_data

        for header_item in header_items:
            directive = header_item.getDirective()
            if directive == STATUS:
                result_data.setStatus(int(header_item.getValue()))
            elif directive == ERROR_MESSAGE:
                result_data.setErrorMessage(header_item.getValue())
            else:
                data_value = data[header_item.getFrom():header_item.getTo()+1]
                if directive == STATIONARY_DISTRIBUTION:
                    result_data.setStatDist(data_value)
                elif directive == TRAJECTORY_PROBABILITY:
                    result_data.setProbTraj(data_value)
                elif directive == TRAJECTORIES:
                    result_data.setTraj(data_value)
                elif directive == FIXED_POINTS:
                    result_data.setFP(data_value)
                elif directive == RUN_LOG:
                    result_data.setRunLog(data_value)
                else:
                    result_data.setErrorMessage("unknown directive " + directive)
                    result_data.setStatus(4)
                    return result_data

        return result_data

    @staticmethod
    def _parse_header_items(header, header_items):
        opos = 0
        pos = 0
        while True:
            pos = header.find(':', opos)
            if pos < 0:
                break
            directive = header[opos:pos+1]
            opos = pos+1
            pos = header.find("\n", opos)
            if pos < 0:
                return "newline not found in header after directive " + directive

            value = header[opos:pos]
            opos = pos+1
            pos2 = value.find("-")
            if directive == STATUS or directive == ERROR_MESSAGE:
                header_items.append(HeaderItem(directive = directive, value = value))
            elif pos2 >= 0:
                header_items.append(HeaderItem(directive = directive, _from = int(value[0:pos2]), to = int(value[pos2+1:])))
            else:
                return "dash - not found in value " + value + " after directive " + directive

        return ""

    @staticmethod
    def _add_header(header, directive, o_offset, offset):
        if o_offset != offset:
            header += directive + str(o_offset) + "-" + str(offset-1) + "\n"

        return (header, offset)



#!/usr/bin/env python

from sys import argv
from maboss import load, UpdatePopulation
from os.path import exists, splitext

if __name__ == "__main__":

    if len(argv) < 4:
        print("Usage:\n")
        print("UpMaBoSS.py <file.bnd> <file.cfg> <file.upp>")
        exit(1)

    bnd_file = argv[1]
    cfg_file = argv[2]
    upp_file = argv[3]

    if not exists(bnd_file):
        print("Cannot find .bnd file")
        exit(1)

    if not exists(cfg_file):
        print("Cannot find .cfg file")
        exit(1)

    if not exists(upp_file):
        print("Cannot find .upp file")
        exit(1)

    workdir = splitext(cfg_file)[0]
    maboss_model = load(bnd_file, cfg_file)
    simulation = UpdatePopulation(maboss_model, upp_file, verbose=True)
    result = simulation.run(workdir=workdir)
    result.save(workdir)


from __future__ import print_function

from ..results.baseresult import BaseResult
from ..results.storedresult import StoredResult
import os
import tempfile
import subprocess
import sys
from random import random
import shutil
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from sklearn.cluster import KMeans
import numpy as np
import multiprocessing
import pandas as pd
import matplotlib.pyplot as plt 
from re import match
import ast

class EnsembleResult(BaseResult):
  
    # def __init__(self, models_files, cfg_filename, prefix="res", individual_results=False, random_sampling=False):
    def __init__(self, simulation, workdir=None, overwrite=False, prefix="res"):

        # self._cfg = cfg_filename
        self.workdir = workdir
        if workdir is None:
            self._path = tempfile.mkdtemp()
        else:
            self._path = workdir
            if os.path.exists(self._path) and overwrite:
                shutil.rmtree(self._path)
                os.mkdir(self._path)
        
            elif not os.path.exists(self._path):
                os.mkdir(self._path)


        self._cfg = os.path.join(self._path, "ensemble.cfg")

        BaseResult.__init__(self, self._path, simulation)
        self.prefix = prefix
        self.asymptotic_probtraj_distribution = None
        self.asymptotic_nodes_probtraj_distribution = None
        maboss_cmd = simulation.get_maboss_cmd()

        simulation.write_cfg(self._path, "ensemble.cfg")
        simulation.write_models(self._path)
        self.models_files = simulation.models_files

        options = ["--ensemble"]
        if simulation.individual_results:
            options.append("--save-individual")

        if simulation.random_sampling:
            options.append("--random-sampling")

        cmd_line = [maboss_cmd] + options

        if len(simulation.individual_cfgs) > 0:
            cmd_line.append("--ensemble-istates")
            for model_file in self.models_files:
                cmd_line += ["-c", os.path.join(self._path, "models", simulation.individual_cfgs[os.path.basename(model_file)])]

        else:
            cmd_line += ["-c", self._cfg]

        cmd_line += [
            "-o", self._path+'/'+self.prefix
        ] + self.models_files

        res = subprocess.Popen(
            cmd_line,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        std_out, std_err = res.communicate()
        self._err = res.returncode
        if self._err != 0:
            print("Error, MaBoSS returned non 0 value", file=sys.stderr)
            print(std_err.decode())
        
        if len(std_out.decode()) > 0:
            print(std_out.decode())

    def get_thread_count(self):
        # TODO : Extracting it from the cfg
        return 6

    def get_fp_file(self):
        return os.path.join(self._path, "%s_fp.csv" % self.prefix)

    def get_probtraj_file(self):
        return os.path.join(self._path, "%s_probtraj.csv" % self.prefix)
        
    def get_statdist_file(self):
        return os.path.join(self._path, "%s_statdist.csv" % self.prefix)

    def load_individual_result(self, model):
        return StoredResult(self._path, self.prefix + "_model_" + str(model))

    def __del__(self):
        if self.workdir is None and os.path.exists(self._path):
            shutil.rmtree(self._path)

    def get_individual_states_probtraj(self, filter=None, cluster=None):
        if self.asymptotic_probtraj_distribution is None:
            results = []
            for i, model in enumerate(self.models_files):
                results.append(self.load_individual_result(i))

            tables = []
            with multiprocessing.Pool(processes=self.get_thread_count()) as pool:
                tables = pool.starmap(get_single_individual_states_distribution, [(result, i) for i, result in enumerate(results)])
            self.asymptotic_probtraj_distribution = pd.concat(tables, axis=0, sort=False)
            self.asymptotic_probtraj_distribution.fillna(0, inplace=True)

        if filter is not None:
            return apply_filter(self.asymptotic_probtraj_distribution, filter, state=True)

        if cluster is not None:
            return self.asymptotic_probtraj_distribution.iloc[cluster, :]
        
        return self.asymptotic_probtraj_distribution

    def get_individual_nodes_probtraj(self, filter=None, cluster=None):
        if self.asymptotic_nodes_probtraj_distribution is None:

            table = self.get_individual_states_probtraj()
            nodes = get_nodes(table.columns.values)
            with multiprocessing.Pool(processes=self.get_thread_count()) as pool:
                self.asymptotic_nodes_probtraj_distribution = pd.concat(
                    pool.starmap(get_single_individual_nodes_distribution, [(table, t_index, nodes) for t_index in table.index]), 
                    sort=False, axis=0
                )

        if filter is not None:
            return apply_filter(self.asymptotic_nodes_probtraj_distribution, filter)

        if cluster is not None:
            return self.asymptotic_nodes_probtraj_distribution.iloc[cluster, :]
        
        return self.asymptotic_nodes_probtraj_distribution

    def getByCondition(self, node_filter=None, state_filter=None):
        if node_filter is not None:
            indexes = self.get_individual_nodes_probtraj(node_filter).index.values
            labels = [0 if i not in indexes else 1 for i in range(len(self.models_files))]
            return indexes, labels
        elif state_filter is not None:
            indexes = self.get_individual_states_probtraj(state_filter).index.values
            labels = [0 if i not in indexes else 1 for i in range(len(self.models_files))]
            return indexes, labels
        

    def filterEnsembleByCondition(self, output_directory, node_filter=None, state_filter=None):

        model_list = None
        if node_filter is not None:
            model_list = self.get_individual_nodes_probtraj(node_filter).index.values
              
        elif state_filter is not None:
            model_list = self.get_individual_states_probtraj(state_filter).index.values
              
        else:
            return None

        if not os.path.exists(output_directory):
            os.mkdir(output_directory)
        else:
            shutil.rmtree(output_directory)
            os.mkdir(output_directory)

        for model in model_list:
            shutil.copyfile(
                self.models_files[int(model)], 
                os.path.join(output_directory, os.path.basename(self.models_files[int(model)]))
            )
    
    def getKMeans(self, clusters=0):
        if clusters > 0:
            kmeans = KMeans(n_clusters=clusters).fit(self.get_individual_nodes_probtraj().values)
            indices = {}
            for i, label in enumerate(kmeans.labels_):
                if label in indices.keys():
                    array = indices[label]
                    array.append(i)
                    indices.update({label: array})
                else:
                    indices.update({label: [i]})

            return indices, kmeans.labels_

    def getStatesKMeans(self, clusters=0):
        if clusters > 0:
            kmeans = KMeans(n_clusters=clusters).fit(self.get_individual_states_probtraj().values)
            indices = {}
            for i, label in enumerate(kmeans.labels_):
                if label in indices.keys():
                    array = indices[label]
                    array.append(i)
                    indices.update({label: array})
                else:
                    indices.update({label: [i]})

            return indices, kmeans.labels_

    def filterEnsembleByCluster(self, output_directory, cluster):

        if cluster is not None:
            if not os.path.exists(output_directory):
                os.mkdir(output_directory)
            else:
                shutil.rmtree(output_directory)
                os.mkdir(output_directory)

            for model in cluster:
                shutil.copyfile(
                    self.models_files[model], 
                    os.path.join(output_directory, os.path.basename(self.models_files[model]))
                )

    def plotSteadyStatesDistribution(self, figsize=None, labels=None, alpha=1, **args):

        pca = PCA()
        table = self.get_individual_states_probtraj()
        mat = table.values
        pca_res = pca.fit(mat)
        X_pca = pca.transform(mat)
        arrows_raw = (np.transpose(pca_res.components_[0:2, :]))
        self.plotPCA(pca, X_pca, list(table.columns.values), list(table.index.values), labels, alpha, figsize=figsize, **args)

    def plotSteadyStatesNodesDistribution(self, compare=None, labels=None, alpha=1, **args):

        pca = PCA()
        table = self.get_individual_nodes_probtraj()
        mat = table.values
        pca_res = pca.fit(mat)
        X_pca = pca.transform(mat)
        arrows_raw = (np.transpose(pca_res.components_[0:2, :]))

        if compare is not None:
            compare_table = compare.get_individual_nodes_probtraj()
            c_pca = pca.transform(compare_table.values)
            self.plotPCA(
                pca, X_pca,
                list(table.columns.values), list(table.index.values), labels, alpha,
                compare=c_pca, **args
            )
        else:
            self.plotPCA(
                pca, X_pca,
                list(table.columns.values), list(table.index.values), labels, alpha,
                **args
            )

    def plotPCA(self, pca, X_pca, samples, features, colors=None, alpha=1, compare=None, figsize=(20, 12), show_samples=False, show_features=True):
        fig = plt.figure(figsize=figsize)

        if colors is None:
            plt.scatter(X_pca[:, 0], X_pca[:, 1], alpha=alpha)
        else:
            legend = ["Cluster #%d" % (i + 1) for i in colors]

            c_colors = ["C%d" % color for color in colors]
            
            scatter = plt.scatter(X_pca[:, 0], X_pca[:, 1], c=c_colors, s=50, alpha=alpha, label=colors)
            import matplotlib.patches as mpatches
            # build the legend

            patches = []
            for cluster in set(colors):
                patches.append(
                    mpatches.Patch(color="C%d" % list(set(colors))[cluster], label='Cluster #%d' % (cluster + 1))
                )

            legend = plt.legend(handles=patches)

        if compare is not None:
            plt.scatter(compare[:, 0], compare[:, 1], alpha=alpha)

        plt.xlabel("PC{} ({}%)".format(1, round(pca.explained_variance_ratio_[0] * 100, 2)))
        plt.ylabel("PC{} ({}%)".format(2, round(pca.explained_variance_ratio_[1] * 100, 2)))
                
        arrows_raw = pca.components_[0:2, :].T
        
        max_x_arrows = max(arrows_raw[:, 0])
        min_x_arrows = min(arrows_raw[:, 0])

        if compare is None:
            max_x_values = max(X_pca[:, 0])
            min_x_values = min(X_pca[:, 0])
        else:
            max_x_values = max(max(X_pca[:, 0]), max(compare[:, 0]))
            min_x_values = min(min(X_pca[:, 0]), min(compare[:, 0]))
        
        max_y_arrows = max(arrows_raw[:, 1])
        min_y_arrows = min(arrows_raw[:, 1])

        if compare is None:
            max_y_values = max(X_pca[:, 1])
            min_y_values = min(X_pca[:, 1])
        else:
            max_y_values = max(max(X_pca[:, 1]), max(compare[:, 1]))
            min_y_values = min(min(X_pca[:, 1]), min(compare[:, 1]))
  
        if show_samples:
            for i, txt in enumerate(features):
                plt.annotate(txt, (X_pca[i, 0], X_pca[i, 1]))
  
        if show_features:
            for i, v in enumerate(arrows_raw):
                plt.arrow(0, 0, v[0], v[1], linewidth=2, color='red')
                plt.text(v[0], v[1], samples[i], color='black', ha='right', va='top', fontsize=18)

            plt.xlim(min(min_x_values, min_x_arrows)*1.2, max(max_x_values, max_x_arrows)*1.2)
            plt.ylim(min(min_y_values, min_y_arrows)*1.2, max(max_y_values, max_y_arrows)*1.2)
        else:
            plt.xlim(min_x_values*1.2, max_x_values*1.2)
            plt.ylim(min_y_values*1.2, max_y_values*1.2)

    def plotTSNESteadyStatesNodesDistribution(self, node_filter=None, state_filter=None, clusters={}, perplexity=50, n_iter=2000, **args):

        pca = PCA()
        table = self.get_individual_nodes_probtraj()
        
        model = TSNE(perplexity=perplexity, n_iter=n_iter, n_iter_without_progress=n_iter*0.5)   
        res = model.fit_transform(table.values)

        if node_filter is None and state_filter is None:
            if len(clusters) == 0:
                fig = plt.figure(**args)
                plt.scatter(res[:, 0], res[:, 1])
            else:
                fig = plt.figure(**args)
                for i, cluster in clusters.items():
                    plt.scatter(res[cluster, 0], res[cluster, 1], color="C%d" % i)

        else:
            fig = plt.figure(**args)
            filtered, _ = self.getByCondition(node_filter=node_filter, state_filter=state_filter)
            not_filtered = list(set(range(len(self.models_files))).difference(set(filtered)))
            
            plt.scatter(res[filtered, 0], res[filtered, 1], color='r')
            plt.scatter(res[not_filtered, 0], res[not_filtered, 1], color='b')

    def plotTSNESteadyStatesDistribution(self, node_filter=None, state_filter=None, clusters={}, perplexity=50, n_iter=2000, **args):

        pca = PCA()
        table = self.get_individual_states_probtraj()
        
        model = TSNE(perplexity=perplexity, n_iter=n_iter, n_iter_without_progress=n_iter*0.5)   
        res = model.fit_transform(table.values)



        if node_filter is None or state_filter is None:
            if len(clusters) == 0:
                fig = plt.figure(**args)
                plt.scatter(res[:, 0], res[:, 1])
            else:
                fig = plt.figure(**args)
                for i, cluster in clusters.items():
                    plt.scatter(res[cluster, 0], res[cluster, 1], color="C%d" % i)

        else:
            fig = plt.figure(**args)
            filtered, _ = self.getByCondition(node_filter=node_filter, state_filter=state_filter)
            not_filtered = list(set(range(len(self.models_files))).difference(set(filtered)))
            
            plt.scatter(res[filtered, 0], res[filtered, 1], color='r')
            plt.scatter(res[not_filtered, 0], res[not_filtered, 1], color='b')


def fix_order(string):
    return " -- ".join(sorted(string.split(" -- ")))

def get_single_individual_states_distribution(result, i):
    if os.path.getsize(result.get_probtraj_file()) > 0:
        raw_table_states = result.get_last_states_probtraj()
        table_states = result.get_last_states_probtraj()
        table_states.rename(index={table_states.index[0]: i}, inplace=True)
        rename_columns = {col: fix_order(col) for col in table_states.columns}
        table_states.rename(columns=rename_columns, inplace=True)
        return table_states

def get_nodes(states):
    nodes = set()
    for s in states:
        if s != '<nil>':
            nds = s.split(' -- ')
            for nd in nds:
                nodes.add(nd)
    return nodes

def get_single_individual_nodes_distribution(table, index, nodes):
    ntable = pd.DataFrame(np.zeros((1, len(nodes))), index=[index], columns=nodes)
    for i, row in enumerate(table):
        state = table.columns[i]
        if state != "<nil>":
            t_nodes = state.split(" -- ")
            for node in t_nodes:
                ntable.loc[index, node] += table.loc[index, state]
                
    return ntable

def apply_filter(data, filter, state=False):

    if state: 
        filter = filter.replace(" -- ", "")
        dict_states = {column.replace(" -- ", ""): column for column in list(data.columns)}
        formula = ast.parse(filter)
        return parse_ast(formula.body[0].value, data, dict_states)
    else:
        formula = ast.parse(filter)
        return parse_ast(formula.body[0].value, data)

def parse_ast(t_ast, data, ds=None):
    if isinstance(t_ast, ast.BoolOp):
        
        if isinstance(t_ast.op, ast.And):
            
            values = [parse_ast(tt_ast, data, ds) for tt_ast in t_ast.values]
            t_data = pd.merge(values[0], values[1], how="inner", on=list(values[0].columns), left_index=True, right_index=True)

            for i in range(2, len(values)):
                t_data = pd.merge(t_data, values[i], how="inner", on=list(values[0].columns), left_index=True, right_index=True)

            return t_data
        elif isinstance(t_ast.op, ast.Or):
            
            values = [parse_ast(tt_ast, data, ds) for tt_ast in t_ast.values]
            t_data = pd.merge(values[0], values[1], how="outer", on=list(values[0].columns), left_index=True, right_index=True)

            for i in range(2, len(values)):
                t_data = pd.merge(t_data, values[i], how="outer", on=list(values[0].columns), left_index=True, right_index=True)

            return t_data
        
    elif isinstance(t_ast, ast.Compare):

        res_dict = {
            ast.Lt : data[parse_ast(t_ast.left, data, ds) < parse_ast(t_ast.comparators[0], data, ds)],
            ast.Gt : data[parse_ast(t_ast.left, data, ds) > parse_ast(t_ast.comparators[0], data, ds)],
            ast.LtE : data[parse_ast(t_ast.left, data, ds) <= parse_ast(t_ast.comparators[0], data, ds)],
            ast.GtE : data[parse_ast(t_ast.left, data, ds) >= parse_ast(t_ast.comparators[0], data, ds)],
            ast.Eq : data[parse_ast(t_ast.left, data, ds) == parse_ast(t_ast.comparators[0], data, ds)],
            ast.NotEq : data[parse_ast(t_ast.left, data, ds) != parse_ast(t_ast.comparators[0], data, ds)],
        }
        return res_dict[type(t_ast.ops[0])]

    elif isinstance(t_ast, ast.Num):
        return t_ast.n

    elif isinstance(t_ast, ast.Name):
        if ds is not None:
            return data[ds[t_ast.id]]
        else:
            return data[t_ast.id]
    
        
        
            


#!/usr/bin/env python
from __future__ import print_function
import os, sys, maboss

def main(argv):

	if len(argv) < 3 or len(argv) > 6:
		print(
			("Bad arguments, MBSS_FormatTable.py <file.bnd> <file.cfg> "
			"<(optional)threshold> <(optional)-mb MaBoSS executable name>"), 
			file=sys.stderr
		)
		exit()

	bndFile = argv[1]
	cfgFile = argv[2]
	mabossCommand = None
	probCutoff = None
	
	if len(argv) > 3:
		index = 3
		while index < len(argv):
			if argv[index].startswith("-mb"):
				mabossCommand = argv[index+1]
				index += 2
			else:
				probCutoff = float(argv[index])
				index += 1

	simulationName = os.path.splitext(os.path.basename(argv[2]))[0]
	
	simulation = maboss.load(bndFile, cfgFile, command=mabossCommand)
	result = simulation.run(command=mabossCommand, prefix=simulationName)
	
	result.save(simulationName)
	
	table = result.get_states_probtraj_full()
	table.to_csv(os.path.join(simulationName, ("%s_probtraj_table.csv" % simulationName)), sep="\t")
	
	result.write_statdist_table(
		os.path.join(simulationName, "%s_statdist_table.csv" % simulationName), 
		probCutoff
	)
	
if __name__ == '__main__':
    main(sys.argv)


"""
Class that contains the cMaBoSS simulation.
"""
from .cmabossresult2 import CMaBoSSResult2

class CMaBoSSSimulation(object):

    def __init__(self, bnd, cfgs):

        self.bnd = bnd
        self.cfgs = cfgs

        self.nb_nodes = self.count_nodes()
        self.cmaboss = self.get_cmaboss()

        self.cmaboss_net = self.cmaboss.MaBoSSNet(self.bnd)
        self.cmaboss_cfg = self.cmaboss.MaBoSSCfg(self.cmaboss_net, self.cfgs)
        
    def count_nodes(self):
        
        res = 0
        with open(self.bnd, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if line.lower().startswith("node"):
                    res += 1
        return res


    def get_cmaboss(self):

        assert self.nb_nodes <= 1024, "Models with more than 1024 nodes are not compatible with this version of MaBoSS"

        if self.nb_nodes <= 64:
            return __import__("cmaboss")
        elif self.nb_nodes <= 128:
            return __import__("cmaboss_128n")
        elif self.nb_nodes <= 256:
            return __import__("cmaboss_256n")
        elif self.nb_nodes <= 512:
            return __import__("cmaboss_512n")
        else:
            return __import__("cmaboss_1024n")

    def run(self, workdir=None, only_final_state=False):
        return CMaBoSSResult2(self, workdir, only_final_state)


__all__ = ["CMaBoSSSimulation"]



"""
Class that contains the stationnary distribution results of a MaBoSS simulation.
"""

import pandas as pd
import numpy as np


class StatDistResult(object):
    
    def __init__(self):
            
        self.state_statdist = None
        self.statdist_clusters = None
        self.statdist_clusters_summary = None
        self.statdist_clusters_summary_error = None

        self._traj_count = None
        self._raw_statdist = None
        self._raw_statdist_clusters = None
        self._raw_statdist_clusters_summary = None
       
    def get_states_statdist(self):
        if self.state_statdist is None:
            raw_statdist, _ = self._get_raw_statdist()
            raw_values = [[val for val in t_data.strip("\n").split("\t")] for t_data in raw_statdist[1:]]
            raw_states = [vals[1::2] for vals in raw_values]
            raw_probas = [vals[2::2] for vals in raw_values]
           
            states = set()
            for t_states in raw_states:
                states.update(t_states)
            states = list(states)
            states_indexes = {state:index for index, state in enumerate(states)}


            indexes = [vals[0] for vals in raw_values]
            new_data = np.zeros((len(raw_states), len(states)))
            
            for i, t_probas in enumerate(raw_probas):
                for j, proba in enumerate(t_probas):
                    new_data[i, states_indexes[raw_states[i][j]]] = np.float64(proba)
            
            self.state_statdist = pd.DataFrame(
                data=new_data,
                columns=states,
                index=indexes

            )
            self.state_statdist.sort_index(axis=1, inplace=True)

        return self.state_statdist


    def get_statdist_clusters(self):
        if self.statdist_clusters is None:
            clusters, _ = self._get_raw_statdist_clusters()
            self.statdist_clusters = []
            for cluster in clusters:
                raw_values = [raw_value.strip("\n").split("\t") for raw_value in cluster]
                raw_states = [values[1::2] for values in raw_values]
                raw_probas = [values[2::2] for values in raw_values]

                states = set()
                for t_states in raw_states:
                    states.update(t_states)
                states = list(states)
                states_indexes = {state:index for index, state in enumerate(states)}

                indexes = [vals[0] for vals in raw_values]
                new_data = np.zeros((len(raw_states), len(states)))
                
                for i, t_probas in enumerate(raw_probas):
                    for j, proba in enumerate(t_probas):
                        new_data[i, states_indexes[raw_states[i][j]]] = np.float64(proba)
                
                df_cluster = pd.DataFrame(
                    data=new_data,
                    columns=states,
                    index=indexes

                )
                df_cluster.sort_index(axis=1, inplace=True)
                self.statdist_clusters.append(df_cluster)

        
        return self.statdist_clusters

    def get_statdist_clusters_summary(self):

        if self.statdist_clusters_summary is None:
            _, summary = self._get_raw_statdist_clusters()
            raw_values = [cluster.strip("\n").split("\t") for cluster in summary]
            raw_states = [values[1::3] for values in raw_values]
            raw_probas = [values[2::3] for values in raw_values]

            states = set()
            for t_states in raw_states:
                states.update(t_states)
            states = list(states)
            states_indexes = {state:index for index, state in enumerate(states)}

            indexes = [vals[0] for vals in raw_values]
            new_data = np.zeros((len(raw_states), len(states)))

            for i, t_probas in enumerate(raw_probas):
                for j, proba in enumerate(t_probas):
                    new_data[i, states_indexes[raw_states[i][j]]] = np.float64(proba)

            self.statdist_clusters_summary = pd.DataFrame(
                data=new_data,
                columns=states,
                index=indexes

            )
            self.statdist_clusters_summary.sort_index(axis=1, inplace=True) 

        return self.statdist_clusters_summary

    def get_statdist_clusters_summary_error(self):

        if self.statdist_clusters_summary_error is None:
            _, summary = self._get_raw_statdist_clusters()
            raw_values = [cluster.strip("\n").split("\t") for cluster in summary]
            raw_states = [values[1::3] for values in raw_values]
            raw_errors = [values[3::3] for values in raw_values]

            states = set()
            for t_states in raw_states:
                states.update(t_states)
            states = list(states)
            states_indexes = {state:index for index, state in enumerate(states)}

            indexes = [vals[0] for vals in raw_values]
            new_data = np.zeros((len(raw_states), len(states)))

            for i, t_errors in enumerate(raw_errors):
                for j, error in enumerate(t_errors):
                    new_data[i, states.index(raw_states[i][j])] = np.float64(error)

            self.statdist_clusters_summary_error = pd.DataFrame(
                data=new_data,
                columns=states,
                index=indexes
            )
            self.statdist_clusters_summary_error.sort_index(axis=1, inplace=True) 

        return self.statdist_clusters_summary_error

    def write_statdist_table(self, filename, prob_cutoff=None):
        clusters, summary = self._get_raw_statdist_clusters()
        with open(filename, 'w') as statdist_table:
            for i_cluster, cluster in enumerate(summary):

                raw_values = cluster.strip("\n").split("\t")
                
                line_0 = "Probability threshold=%s\n" % (str(prob_cutoff) if prob_cutoff is not None else "") 
                line_1 = ["Prob[Cluster %s]" % raw_values[0]] 
                line_2 = ["%g" % (len(clusters[i_cluster])/self._traj_count)]
                line_3 = ["ErrorProb"]

                for i in range(0, len(raw_values[1:]), 3):
                    if prob_cutoff is None or raw_values[1:i+1] > prob_cutoff:
                        line_1.append("Prob[%s | Cluster %s]" % (raw_values[1+i], raw_values[0]))
                        line_2.append("%s" % raw_values[1+i+1])
                        line_3.append("%s" % raw_values[1+i+2])

                statdist_table.write(line_0)
                statdist_table.write("\t".join(line_1) + "\n")
                statdist_table.write("\t".join(line_2) + "\n")
                statdist_table.write("\t".join(line_3) + "\n")
                statdist_table.write("\n")

    def _get_statdist_fd(self):
        return open(self.get_statdist_file(), "r")

    def _get_raw_statdist(self):
        if self._raw_statdist is None or self._traj_count is None:
            with self._get_statdist_fd() as f:
                self._raw_statdist = []
                self._traj_count = None
                for i, line in enumerate(f.readlines()):
                    if self._traj_count is None:
                        if len(line.strip("\n")) == 0:
                            self._traj_count = i
                            break
                        
                        self._raw_statdist.append(line)

        return self._raw_statdist, self._traj_count

    def _get_raw_statdist_clusters(self):
        if self._raw_statdist_clusters is None or self._raw_statdist_clusters_summary is None:
                
            _, traj_count = self._get_raw_statdist()
            self._raw_statdist_clusters = []
            self._raw_statdist_clusters_summary = []

            with self._get_statdist_fd() as f:
                started_cluster = False
                t_cluster = []
                started_summary = False
                for i, line in enumerate(f.readlines()[self._traj_count+1:]):
                    if not started_summary:
                        if len(line.strip("\n")) == 0:
                            self._raw_statdist_clusters.append(t_cluster)
                            t_cluster = []
                            started_cluster = False

                        elif line.startswith("Cluster\tState"):
                            started_summary = True
                        elif not started_cluster:
                            started_cluster = True
                        else:
                            t_cluster.append(line)

                    else:
                        self._raw_statdist_clusters_summary.append(line)

        return self._raw_statdist_clusters, self._raw_statdist_clusters_summary




from __future__ import print_function
import os
import sys
import re
import random
import pandas as pd
import numpy as np
if sys.version_info[0] < 3:
    from contextlib2 import ExitStack
else:
    from contextlib import ExitStack
import glob
from ..results.storedresult import StoredResult
from ..server import MaBoSSClient
import shutil
from multiprocessing import Pool
from collections import OrderedDict


class CMaBoSSUpdatePopulationResults:
    def __init__(self, uppModel, verbose=False, workdir=None, overwrite=False, previous_run=None, previous_run_step=-1, nodes_init=None, only_final_state=False):
        """UpdatePopulationResults class
        :param uppModel: UppMaBoSS model
        :param verbose: boolean to activate verbose mode, default to False        
        :param previous_run: previous run used to initialize starting probability
        distribution, default to None
        :param previous_run_step: step used to initialize starting probability
        distribution from previous run, default to -1 (last)
        """
        self.uppModel = uppModel
        self.pop_ratios = pd.Series()
        self.stepwise_probability_distribution = None
        self.nodes_stepwise_probability_distribution = None
        self.nodes_list_stepwise_probability_distribution = None

        self.workdir = workdir
        self.overwrite = overwrite
        self.verbose = verbose
        self.nodes_init = nodes_init
        self.only_final_state = only_final_state
        self.results = []
        self.pop_ratio = uppModel.pop_ratio
 
        if self.workdir is not None and os.path.exists(self.workdir) and not self.overwrite:
            # Restoring
            self.results = [None] * (self.uppModel.step_number + 1)

            for folder in sorted(glob.glob("%s/Step_*_probtraj.csv/" % self.workdir)):
                step = os.path.basename(folder[0:-1]).split("_")[-1]
                self.results[int(step)] = StoredResult(folder)

            for folder in sorted(glob.glob("%s/Step_*_finalprob.csv/" % self.workdir)):
                step = os.path.basename(folder[0:-1]).split("_")[-1]
                self.results[int(step)] = StoredResult(folder)

            self.pop_ratios = pd.read_csv(
                os.path.join(self.workdir, "PopRatios.csv"),
                index_col=0, squeeze=True
            ) / self.uppModel.base_ratio

            if previous_run:
                # Load the previous run final state
                _get_next_condition_from_trajectory(previous_run, self.uppModel.model, step=previous_run_step)

        else:
            if workdir is not None:
                if os.path.exists(workdir):
                    shutil.rmtree(workdir)
                os.makedirs(workdir)

            if previous_run:
                # Load the previous run final state
                _get_next_condition_from_trajectory(previous_run, self.uppModel.model, step=previous_run_step)

            self._run()

    def _run(self):
        
        if self.verbose:
            print("Run MaBoSS step 0")
        
        sim_workdir = os.path.join(self.workdir, "Step_0") if self.workdir is not None else None

        result = self.uppModel.model.run(workdir=sim_workdir, cmaboss=True, only_final_state=self.only_final_state)
        
        self.results.append(result)
        self.pop_ratios[self.uppModel.time_shift] = self.pop_ratio
    
        modelStep = self.uppModel.model.copy()

        for stepIndex in range(1, self.uppModel.step_number+1):
            #
            # Update pop ratio and construct the new version of model
            #
            modelStep = self._buildUpdateCfg(modelStep, result, stepIndex)
            
            if modelStep is None:
                if self.verbose:
                    print("No cells left")

                break

            else:
                if self.verbose:
                    print("Running MaBoSS for step %d" % stepIndex)

                sim_workdir = os.path.join(self.workdir, "Step_%d" % stepIndex) if self.workdir is not None else None

                result = modelStep.run(workdir=sim_workdir, cmaboss=True, only_final_state=self.only_final_state)
                
                self.results.append(result)

        if self.workdir is not None:
            self.save_population_ratios(os.path.join(self.workdir, "PopRatios.csv"))

    def get_population_ratios(self, name=None):
        if name:
            self.pop_ratios.name = name
        return self.pop_ratios*self.uppModel.base_ratio

    def get_stepwise_probability_distribution(self, nb_cores=1, include=None, exclude=None):
        if self.stepwise_probability_distribution is None:
            if nb_cores > 1:
                tables = []
                with Pool(processes=nb_cores) as pool:
                    tables = pool.map(
                        make_stepwise_probability_distribution_line, self.results
                    )

            else:
                tables = [result.get_last_states_probtraj() for result in self.results]
            
            self.stepwise_probability_distribution = pd.concat(tables, axis=0, sort=False)
            self.stepwise_probability_distribution.fillna(0, inplace=True)
            self.stepwise_probability_distribution.set_index([list(range(0, len(tables)))], inplace=True)
            self.stepwise_probability_distribution.insert(
                0, column='PopRatio', value=(self.pop_ratios*self.uppModel.base_ratio).values
            )
            
        if include is None and include is None:
            return self.stepwise_probability_distribution
        else:
            states_filtered = self.stepwise_probability_distribution.columns
            
            if include is not None:
                states_filtered = [state for state in states_filtered if set(include).issubset(set(state.split(" -- ")))]
    
            if exclude is not None:
                states_filtered = [state for state in states_filtered if set(exclude).isdisjoint(set(state.split(" -- ")))]
           
            return self.stepwise_probability_distribution.loc[:, states_filtered]

    def get_nodes_stepwise_probability_distribution(self, nodes=None, nb_cores=1):
        if self.nodes_stepwise_probability_distribution is None or set(nodes) != self.nodes_list_stepwise_probability_distribution:
            
            table = self.get_stepwise_probability_distribution(nb_cores=nb_cores)
            
            states = table.columns.values[1:].tolist()
            if "<nil>" in states:
                states.remove("<nil>")

            if nodes is None:
                nodes = get_nodes(states)
            else:
                nodes = set(nodes)
            
            self.nodes_list_stepwise_probability_distribution = nodes

            node_dict = {}
            for state in states:
                t_nodes = state.split(" -- ")
                t_nodes = [node for node in t_nodes if node in nodes]
                if len(t_nodes) > 0:
                    node_dict.update({state: t_nodes})

            if nb_cores > 1:
                self.nodes_stepwise_probability_distribution = make_nodes_table_parallel(table, nodes, node_dict, nb_cores)
            else:
                self.nodes_stepwise_probability_distribution = make_nodes_table(table, nodes, node_dict)

            self.nodes_stepwise_probability_distribution.insert(0, column='PopRatio', value=(self.pop_ratios*self.uppModel.base_ratio).values)

        return self.nodes_stepwise_probability_distribution

    def save(self, path):
        if not os.path.exists(path):
            os.mkdir(path)

        self.save_model(path)
        self.save_population_ratios(os.path.join(path, "PopRatios.csv"))
        self.save_stepwise_probability_distribution(os.path.join(path, "PopProbTraj.csv"))

    def save_model(self, path):
        with ExitStack() as stack:
            cfg_file = stack.enter_context(open(os.path.join(path, "model.cfg"), 'w'))
            bnd_file = stack.enter_context(open(os.path.join(path, "model.bnd"), 'w'))
            self.uppModel.model.print_cfg(cfg_file)
            self.uppModel.model.print_bnd(bnd_file)

    def save_population_ratios(self, path):
        (self.pop_ratios*self.uppModel.base_ratio).to_csv(path, header=["PopRatio"], index_label="Step")

    def save_stepwise_probability_distribution(self, path):
        nb_cores = int(self.uppModel.model.param["thread_count"])
        self.get_stepwise_probability_distribution(nb_cores=nb_cores).to_csv(path, index_label="Step")

    def _buildUpdateCfg(self, simulation, result, stepIndex):
        """Configure the MaBoSS model for the next run of UppMaBoss.
        In practice, _buildUpdateCfg uses the previous simulation result
        to compute pop ratio, death, division, parameters, nodes formulas 
        and sets the init state of the model for the next step of simulation
        :param simulation: MaBoSS simulation
        :param result: result object of the previous run        
        :param stepIndex: current step of the UppMaBoss simulation
        """
        #
        # Read first and last line, extract last states with respective probs
        #
        # first_line, last_line = read_first_last_lines_from_trajectory (traj_fd)
        # states, probs = get_states_probs_from_trajectory_line (first_line, last_line)
        last_state, raw_states, timepoint = result.cmaboss_result.get_last_probtraj()
        
        states = []
        probs = []
        for i_state, prob in enumerate(last_state[0]):
            states.append(raw_states[i_state].split(" -- "))
            probs.append(prob)
        #
        # Update pop ratio
        #
        self.pop_ratio *= self._updatePopRatio (states, probs)
        new_time = self.uppModel.time_shift + self.uppModel.time_step*stepIndex
        self.pop_ratios[new_time] = self.pop_ratio
        #
        # Normalize
        #
        states, probs = self.normalize_with_death_and_division (states, probs)
        if states is None:
            return None        
        #
        # Compute formulas for parameters and nodes 
        # 
        parameters = self.compute_parameters(simulation, states, probs)
        nodes_with_formula = self.compute_nodes_formula (states, probs)
        #
        # Apply new values for parameters 
        # 
        simulation.param.update(parameters)
        #
        # Init states
        #
        nodes_to_init, new_istate = self._initCond_Trajline (states, probs)
        simulation.network.set_istate (nodes_to_init, new_istate, warnings=False)
        #
        # Init nodes having a formula
        #
        for a_node in nodes_with_formula.keys():
            new_val = nodes_with_formula[a_node]
            simulation.network.set_istate(a_node, [1-new_val,new_val], warnings=False)
        return simulation 
    
    def normalize_with_death_and_division(self, states, probs): 
        """
        Take into account impact of death and division and normalize
        NB: if no death, nor division is defined, do nothing
        :param states: list of states extracted from the trajectory
        :param probs: list of states probabilities extracted from the trajectory  
        """
        #
        # speed up programm when no death, nor division
        # 
        if not self.uppModel.death_node and not self.uppModel.division_node:
            return states, probs
        #
        # if death or division
        #
        norm_factor = 0
        death_prob = 0
        division_prob = 0
    
        states_ret = []
        probs_ret = []
        for one_state, one_prob in zip(states, probs):
            one_state_ret = one_state.copy()
            one_prob_ret = one_prob
            
            if self.uppModel.death_node in one_state_ret:
                death_prob += one_prob_ret
                one_prob_ret = 0
            else:
                if self.uppModel.division_node in one_state_ret:
                    division_prob += one_prob_ret
                    one_prob_ret *= 2.0
                    one_state_ret.remove(self.uppModel.division_node)
                norm_factor += one_prob_ret
                
            states_ret.append (one_state_ret)
            probs_ret.append (one_prob_ret)
    
        if self.verbose:
            print("Norm Factor:%g probability of death: %g probability of division: %g"  \
                  % (norm_factor, death_prob, division_prob))
        #
        # if norm_factor <= 0, no more cells
        #
        if norm_factor <= 0:
            return None, None
        #
        # If norm_factor > 0, normalize 
        #
        probs_ret = np.array(probs_ret, dtype = float)
        probs_ret = (probs_ret / norm_factor).tolist()
        return states_ret, probs_ret

    def compute_parameters(self, simulation, states, probs): 
        """
        Computer parameters
        :param simulation: MaBoss model (containing the defining of parameters) 
        :param states: list of states extracted from the trajectory
        :param probs: list of states probabilities extracted from the trajectory  
        """
        parameters = {}
        for parameter, value in simulation.param.items():
            if parameter.startswith("$") and parameter in self.uppModel.update_var.keys():
                formula = self.uppModel.update_var[parameter]
                new_value = varDef_Upp(formula, states, probs)
                for match in re.findall("#rand", new_value):
                    rand_number = random.uniform(0, 1)
                    new_value = new_value.replace("#rand", str(rand_number), 1)
    
                new_value = new_value.replace("#pop_ratio", str(self.pop_ratio))
                parameters.update({parameter: new_value})
                if self.verbose:
                    print("Updated variable: %s = %s" % (parameter, new_value))
        return parameters

    def compute_nodes_formula (self, states, probs): 
        """
        Computer nodes formula to be used as init values for next run
        :param states: list of states extracted from the trajectory
        :param probs: list of states probabilities extracted from the trajectory  
        """
        all_node_upd = {}
        for node_upd in self.uppModel.nodes_formula.keys():
            node_formula = self.uppModel.nodes_formula[node_upd]
            new_value = varDef_Upp(node_formula, states, probs)
            
            for match in re.findall("#rand", new_value):
                rand_number = random.uniform(0, 1)
                new_value = new_value.replace("#rand", str(rand_number), 1)
    
            new_value = new_value.replace("#pop_ratio", str(self.pop_ratio))
            #
            # Remove trailing ';' added by varDef_Upp, compute formula via eval 
            # and convert to proba
            #
            new_value = float (eval(new_value[:-1]))
            new_value = np.clip(new_value, 0, 1)
               
            all_node_upd.update({node_upd: new_value})
            if self.verbose:
                print("Updated node:", node_upd, "=", new_value)
        return all_node_upd

    def _initCond_Trajline(self, states, probs):
        """
        Return the list of states to be initialized from states and probs
        The function excludes from states the nodes with formulas
        or (for the first run) nodes with init value supplied as parameter
        :param states: list of states extracted from the trajectory
        :param probs: list of states probabilities extracted from the trajectory  
        :param nodes_init: dict of nodes values of the form { "NODE1" : TrueValue1, "NODE2" : TrueValue2, ... }.
        Nodes to exclude from InitCond as these nodes have a specific init value
        """
        new_istate = OrderedDict()
        #
        # Remove from the list of nodes the ones having a rule or an init value
        #
        nodes_to_exclude = set(self.uppModel.nodes_formula.keys()) 
        if self.nodes_init:
            nodes_to_exclude= nodes_to_exclude | set(self.nodes_init.keys())

        list_nodes_to_set = list( set(self.uppModel.node_list) - nodes_to_exclude )
        # Not necessary for MaBoss, only to have visually an ouput always in the same order
        list_nodes_to_set.sort()
        #
        # Associate an index to each node
        #
        name2idx = {name: i for i, name in enumerate(list_nodes_to_set)}
        #
        # Exclude nodes with formula or with specific init from states
        #
        states = exclude_nodes_from_states (states, nodes_to_exclude)
        #
        # Construct a list of states (each state is tuple of nodes on/off)
        # with the probability of each state
        #
        for state, prob in zip(states, probs):
            #
            # Inclusion of modified _str2state code
            #
            str2state = [ 0 ] * len(name2idx)
            if '<nil>' not in state:
                for node in state:
                    str2state[name2idx[node]] = 1
            
            state_tuple = tuple(str2state)

            if state_tuple in new_istate.keys():
                prob += new_istate[state_tuple]
            
            new_istate.update({state_tuple: prob})

        return list_nodes_to_set, new_istate

    def _updatePopRatio (self, states, probs):
        """
        Return update population ratio using nodes having death or division
        :param states: states extracted from the trajectory
        :param probs: probabilities of states extracted from the trajectory  
        """
        upd_pop_ratio = 0.0
        for a_state, a_prob in zip(states, probs):
            if self.uppModel.death_node not in a_state:
                if self.uppModel.division_node in a_state:
                    upd_pop_ratio += 2 * a_prob
                else:
                    upd_pop_ratio += a_prob
        return upd_pop_ratio

def varDef_Upp(update_line, states, probs):
	res_match = re.findall("p\[[^\[]*\]", update_line)
	if len(res_match) == 0:
		print("Syntax error in the parameter update definition : %s" % update_line, file=sys.stderr)
		exit()
	
	for match in res_match:

		lhs, rhs = match.split("=")
		lhs = lhs.replace("p[", "").replace("]", "").replace("(", "").replace(")", "")
		rhs = rhs.replace("[", "").replace("]", "").replace("(", "").replace(")", "")

		node_list = [token.strip() for token in lhs.split(",")]
		boolVal_list = [token.strip() for token in rhs.split(",")]

		if len(node_list) != len(boolVal_list):
			print("Wrong probability definitions for \"%s\"" % match)
			exit()

		upNodeList = []
		downNodeList = []

		for i, node in enumerate(node_list):
			if float(boolVal_list[i]) == 0.0:
				downNodeList.append(node)
			else:
				upNodeList.append(node)

		probValue = 0.0
		for i in range(0, len(states)):
			upNodeProbTraj = states[i]
			interlength = 0

			for upNodePt in upNodeProbTraj:
				for upNode in upNodeList:
					if upNodePt == upNode:
						interlength += 1

			if interlength == len(upNodeList):
				interlength = 0
				for upNodePt in upNodeProbTraj:
					for downNode in downNodeList:
						if upNodePt == downNode:
							interlength = 1
							break
					
					if interlength == 1:
						break
				if interlength == 0:
					probValue += probs[i]

		update_line = update_line.replace(match, str(probValue), 1)
	update_line += ";"
	return update_line

def _get_next_condition_from_trajectory(self, next_model, step=-1):
    """
    Set the values of MaBoss model when resuming from a previous run
    :param next_model : MaBoss model
    :param step: step of the UppMaBoss simulation
    """
    #
    # Extract states and probs from trajectory
    #
    last_state = result.cmaboss_result.get_last_states_probtraj()
    states = []
    probs = []
    for state, prob in last_state.items():
        states.append(state.split(" -- "))
        probs.append(prob)
    #
    # Compute formulas for nodes 
    # 
    nodes_with_formula = self.compute_nodes_formula (states, probs)
    #
    # Init states
    #
    nodes_to_set, new_istate = self._initCond_Trajline(states, probs)        
    next_model.network.set_istate (nodes_to_set, new_istate, warnings=False)    
    #
    # Init nodes having a formula
    #
    for a_node in nodes_with_formula.keys():
        new_val = nodes_with_formula[a_node]
        next_model.network.set_istate(a_node, [1-new_val,new_val], warnings=False)
    #
    # Init nodes having an init value
    #
    if self.nodes_init:
        for a_node, new_val in self.nodes_init.items():
            next_model.network.set_istate(a_node, [1-new_val,new_val], warnings=False)
            if self.verbose:
                print("Starting init of node:", a_node, "=", new_val)

def exclude_nodes_from_states (traj_states, nodes_exluded=None):
    """Exclude nodes from states
    :param traj_states: list of states (each state is a liste of nodes)
    :param nodes_exluded: list/set of nodes to remove
    """
    if (not nodes_exluded) or len(nodes_exluded) <= 0:
        return traj_states
    traj_states_ret = []
    for one_traj_states in traj_states:
        new_traj_states = list( set(one_traj_states) - set(nodes_exluded) )
        if len (new_traj_states) <= 0:
            new_traj_states  = ["<nil>"]
        traj_states_ret.append (new_traj_states)
    return traj_states_ret   

def make_stepwise_probability_distribution_line(result):
    return result.get_last_states_probtraj()

def get_nodes(states):
    nodes = set()
    for s in states:
        if s != '<nil>':
            nds = s.split(' -- ')
            for nd in nds:
                nodes.add(nd)
    return nodes

def make_node_line(row, states_table, index, node_dict):
    for state, nd_state in node_dict.items():
        for nd in nd_state:
            row[nd] += states_table.iloc[index, states_table.columns.get_loc(state)]

def make_nodes_table(spd, nodes, node_dict):
    table = pd.DataFrame(
        np.zeros((len(spd.index), len(nodes))),
        index=spd.index.values, columns=nodes
    )

    for index, (_, row) in enumerate(table.iterrows()):
        make_node_line(row, spd, index, node_dict)
        
    return table

def make_node_line_parallel(states_row, nodes, node_dict, row):
    
    table = pd.DataFrame(
        np.zeros((1, len(nodes))),
        index=[row], columns=nodes
    )
    
    for row_index, row_state in states_row.iteritems():
        if row_index in node_dict.keys() and row_state > 0.0:
            for nd in node_dict[row_index]:
                table.loc[row, nd] += row_state
    
    return table

def make_nodes_table_parallel(spd, nodes, node_dict, nb_cores):
    
    tables = []
    with Pool(processes=6) as pool:
        tables = pool.starmap(
            make_node_line_parallel, 
            [(spd.iloc[index, :], nodes, node_dict, row) for index, row in enumerate(spd.index)]
        )

    table = pd.concat(tables, axis=0, sort=False)

    return table
